<?php

namespace Intergo\Wallet;

use DateTime;
use Intergo\Wallet\Enums\BalanceEnum;
use Intergo\Wallet\Exceptions\InvalidAmountException;
use Intergo\Wallet\Exceptions\InvalidBalanceException;
use Intergo\Wallet\ValueObjects\TransactionInfoVO;
use Illuminate\Support\Facades\DB;
use Illuminate\Support\Facades\Redis;

abstract class AbstractWalletService
{
    private Redis $redis;
    private DB $db;

    /**
     * AbstractWalletService constructor.
     */
    public function __construct(Redis $redis, DB $db)
    {
        $this->redis = $redis;
        $this->db = $db;
    }

    abstract public function log($message, $data);

    /**
     * Add Balance to user on Redis
     *
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    public function addBalance($amount, array $information = [])
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO($amount, $information);
        return $this->adjustBalance($transactionInfo);
    }

    /**
     * Deduct Balance to user on Redis
     *
     * @param $amount
     * @param array $information
     * @return float
     * @throws InvalidAmountException
     * @throws InvalidBalanceException
     */
    public function deductBalance($amount, array $information = [])
    {
        $this->validateAmount($amount);
        $transactionInfo = new TransactionInfoVO(-$amount, $information);
        return $this->adjustBalance($transactionInfo);
    }

    /**
     * Either Deduct or Add Balance on Redis + add transaction for the balance change
     *
     * @param TransactionInfoVO $transactionInfo
     * @return float
     * @throws InvalidBalanceException
     */
    protected function adjustBalance(TransactionInfoVO $transactionInfo)
    {
        $userId = $transactionInfo->getUserId();
        $amount = $transactionInfo->getAmount();
        $balance = $transactionInfo->setBalance($this->getBalance($userId));
        $newBalance = $transactionInfo->setNewBalance($balance + $amount);
        if ($newBalance < 0) {
            $this->log('Cash Transactions Discrepancies', $transactionInfo);
            throw new InvalidBalanceException();
        }
        $this->redis::connection(BalanceEnum::BALANCE_REDIS_CONNECTION)->incrbyfloat('User:' . $userId . ':balance', $amount);
        $now = date("Y-m-d H:i:s");
        $transactionArchived = [
            'user_id' => $userId,
            'amount' => $amount,
            'old_cash_balance' => $balance,
            'cash_balance' => $newBalance,
            'reason' => $transactionInfo->getReason(),
            'remarks' => $transactionInfo->getRemarks(),
            'created_at' => $now,
            'updated_at' => $now
        ];
        $this->redis::connection(BalanceEnum::BALANCE_REDIS_CONNECTION)->rpush('Transactions:' . $userId, json_encode($transactionArchived));
        $this->log('Cash Transactions', $transactionInfo);
        return $newBalance;
    }

    /**
     * Set the balance. Instead of directly assigning the value to redis, the difference in amount is added/deducted and maintain the transactions
     *
     * @param TransactionInfoVO $transactionInfo
     * @return float
     * @throws InvalidBalanceException
     */
    public function setBalance(TransactionInfoVO $transactionInfo)
    {
        $currentBalance = $this->getBalance($transactionInfo->getUserId());
        $amount = abs($transactionInfo->getAmount());
        $transactionInfo->setAmount($amount - $currentBalance);
        return $this->adjustBalance($transactionInfo);
    }

    /**
     * Get the balance from redis. If redis doesn't exist, adjust the balance from transactions and store to redis
     *
     * @return float
     */
    public function getBalance($userId)
    {
        $balance = $this->redis::connection(BalanceEnum::BALANCE_REDIS_CONNECTION)->get('User:' . $userId . ':balance');
        if (!$balance) {
            if ($userId) {
                $balance = $this->getBalanceFromDb($userId);
            } else {
                $balance = 0;
            }
        }
        return $this->roundBalance($balance);
    }

    /**
     * Get the balance from DB
     *
     * @param int $userId
     * @return float|int
     */
    protected function getBalanceFromDb(int $userId)
    {
        $transactionInfoVO = new TransactionInfoVO(0);
        $transactionInfoVO->setUserId($userId);
        $user = $this->db::connection('auth_mysql')->table(BalanceEnum::USERS_TABLE)
            ->select('id', 'cash_balance')
            ->where('id', $userId)
            ->first();
        if (!$user) {
            $this->log( 'User not found for cash_balance from DB. Setting balance to 0', $transactionInfoVO);
            return 0;
        }
        $transactionInfoVO->setBalance($user->cash_balance);
        $this->log( 'Fetch user cash_balance from DB', $transactionInfoVO);

        $count = $this->redis::connection(BalanceEnum::BALANCE_REDIS_CONNECTION)->llen('Transactions:' . $userId);
        $transactionInfoVO->setCountTransactions($count);
        $this->log('Count transactions of user from Redis with Key Transactions:' . $userId, $transactionInfoVO);

        $transactions = $this->redis::connection(BalanceEnum::BALANCE_REDIS_CONNECTION)->lrange('Transactions:' . $userId, 0, $count);
        $this->log('Fetch now transactions of user from Redis with Key Transactions:' . $userId, $transactionInfoVO);
        $sum = 0;
        foreach ($transactions as $transaction) {
            $sum += json_decode($transaction)->amount;
        }
        $newBalance = $transactionInfoVO->setNewBalance($user->cash_balance + $sum);
        $this->log('Final balance of the user using Transactions from Redis', $transactionInfoVO);
        $this->redis::connection(BalanceEnum::BALANCE_REDIS_CONNECTION)->set('User:' . $userId . ':balance', $newBalance);
        $this->log('Saved final balance of the user on Redis', $transactionInfoVO);
        return $this->roundBalance($newBalance);
    }

    /**
     * @param $amount
     * @throws InvalidAmountException
     */
    private function validateAmount($amount)
    {
        if (is_null($amount) || is_bool($amount)) {
            throw new InvalidAmountException();
        }
    }

    /**
     * @param $balance
     * @return float
     */
    private function roundBalance($balance)
    {
        return round($balance, BalanceEnum::BALANCE_PRECISION);
    }
}
